################################################################################
#
#  Unit group generators
#
################################################################################

function _unit_group_generators(O)
  M = maximal_order(O)
  gens = _unit_group_generators_maximal(M)
  _, Y = _orbit_stabilizer(gens, one(algebra(O)), O)
  return Y
end

function _unit_group_generators_maximal(M)
  res = decompose(algebra(M))
  Mbas = basis(M)
  idems = [mB(one(B)) for (B, mB) in res]
  gens = elem_type(algebra(M))[]
  for i in 1:length(res)
    B, mB = res[i]
    MinB = Order(B, [(mB\(mB(one(B)) * elem_in_algebra(b))) for b in Mbas])
    UB = _unit_group_generators_maximal_simple(MinB)
    e = sum(idems[j] for j in 1:length(res) if j != i; init = zero(algebra(M)))
    @assert isone(e + mB(one(B)))
    for u in UB
      push!(gens, mB(u) + e)
    end
  end
  @assert all(u in M for u in gens)
  @assert all(inv(u) in M for u in gens)
  return gens
end

function _unit_group_generators_maximal_simple(M)
  A = algebra(M)
  if dim(A) == 0
    return [zero(A)]
  end
  ZA, ZAtoA = _as_algebra_over_center(A)
  if dim(ZA) == 1
    # this is a field
    K, AtoK = _as_field_with_isomorphism(A)
    OK = maximal_order(K)
    u, mu = unit_group(OK)
    return [preimage(AtoK, elem_in_nf(mu(u[i]))) for i in 1:ngens(u)]
  elseif isdefined(A, :isomorphic_full_matrix_algebra)
    B, AtoB = A.isomorphic_full_matrix_algebra
    OB = _get_order_from_gens(B, [AtoB(elem_in_algebra(b)) for b in absolute_basis(M)])
    N, S = nice_order(OB)
    @assert basis_matrix(N) == identity_matrix(base_ring(B), dim(B))
    gens = [ B(u) for u in _GLn_generators(base_ring(OB), degree(B))]
    @assert all(b in N for b in gens)
    gens_adjusted = [ inv(S) * u * S for u in gens]
    @assert all(b in OB for b in gens_adjusted)
    gens_in_M = [ AtoB\u for u in gens_adjusted]
    @assert all(b in M for b in gens_in_M)
    return gens_in_M
  elseif dim(ZA) == 4 && !is_split(ZA) && !isdefined(A, :isomorphic_full_matrix_algebra)
    Q, QtoZA = is_quaternion_algebra(ZA)
    MQ = _get_order_from_gens(Q, [QtoZA\(ZAtoA\(elem_in_algebra(b))) for b in absolute_basis(M)])
    _gens =  _unit_group_generators_quaternion(MQ)
    gens_in_M = [ ZAtoA(QtoZA(elem_in_algebra(u))) for u in _gens]
    @assert all(b in M for b in gens_in_M)
    return gens_in_M
  else
    throw(NotImplemented())
  end
end

function _SLn_generators(OK, n)
  K = nf(OK)
  res = dense_matrix_type(K)[]

  if n == 1
    return dense_matrix_type(K)[identity_matrix(K, 1)]
  end

  if n == 2
    r = unit_group_rank(OK)
    if r > 0
      # https://mathoverflow.net/questions/105857/generators-for-sl-2r-for-rings-of-integers-r
      B = basis(OK)
      for b in B
        push!(res, matrix(K, 2, 2, [1, elem_in_nf(b), 0, 1]))
        push!(res, matrix(K, 2, 2, [1, 0, elem_in_nf(b), 1]))
      end
      return res
    else
      if degree(K) > 1
        throw(NotImplemented())
      else
        # This is the case K = Q
        # SL(2, Z) (the modular group) is generated by S, T
        push!(res, matrix(K, 2, 2, [0, -1, 1, 0]))
        push!(res, matrix(K, 2, 2, [1, 1, 1, 0]))
        return res
      end
    end
  else
    # This is the case n >= 3
    # Follows from Bass, "K-Theory and stable algebra", Corollary 1.5
    d = degree(K)
    #
    # We find a small generating set of OK as Z-algebra
    found = false
    local G::Vector{AbsSimpleNumFieldElem}
    for i in 1:d
      for j in 1:10
        G = [elem_in_nf(rand(OK, 2)) for k in 1:i]
        fl,_ = defines_order(K, G)
        if !fl
          continue
        end
        OO = Order(K, G)
        if abs(discriminant(OO)) == abs(discriminant(OK))
          found = true
          break
        end
      end
      if found
        break
      end
    end

    if !found
      B = elem_in_nf.(basis(OK))
    else
      B = G
      if !(one(OK) in B)
        push!(B, elem_in_nf(one(OK)))
      end
    end

    for i in 1:n
      for j in 1:n
        if j == i
          continue
        end
        for b in B
          M = identity_matrix(K, n)
          M[i, j] = b
          push!(res, M)
        end
      end
    end
    return res
  end
end

function _GLn_generators(OK, n)
  K = nf(OK)
  if degree(K) == 1
    if n == 0
      return dense_matrix_type(K)[]
    elseif n == 1
      return dense_matrix_type(K)[matrix(K, 1, 1, [-1])]
    else
      # https://mathoverflow.net/questions/181366/minimal-number-of-generators-for-gln-mathbbz
      # The number should be reduced once more
      s1 = zero_matrix(K, n, n)
      for i in 2:nrows(s1)
        s1[i, i - 1] = 1
      end
      s1[1, n] = 1

      s3 = identity_matrix(K, n)
      s3[1, 2] = 1

      s4 = identity_matrix(K, n)
      s4[1, 1] = -1
      return [s1, s3, s4]
    end
  end

  if n == 2 && unit_group_rank(OK) == 0 && degree(K) == 2
    return _GLn_generators_quadratic(OK, n)
  end

  res = _SLn_generators(OK, n)
  U, mU = unit_group(OK)
  if n == 0
    return dense_matrix_type(K)[]
  end

  for i in 1:ngens(U)
    I = identity_matrix(K, n)
    I[1, 1] = elem_in_nf(mU(U[i]))
    push!(res, I)
  end

  return res
end

function _GLn_generators_quadratic(OK, n)
  d = discriminant(OK)
  _v = findfirst(z -> z[1] == d, __GLn_generators_quadratic)
  if _v === nothing
    throw(NotImplemented())
  end
  v = __GLn_generators_quadratic[_v]
  dn = v[2]
  L, = quadratic_field(-dn, cached = false)
  K = number_field(OK)
  fl, i = is_isomorphic_with_map(L, K)
  res = dense_matrix_type(K)[]
  for w in v[3]
    mat = matrix(K, 2, 2, [i(L(u)) for u in w])
    @assert is_unit(det(mat))
    push!(res, mat)
  end
  return res
end

global __GLn_generators_quadratic = [(-4, 1, [[[ 1, 0 ],[ 0, 0 ],[ 0, -1 ],[ 1, 0 ]],[[ 1, 0 ],[ 0, 1 ],[ 0, 0 ],[ 0, 1 ]],[[ 0, -1 ],[ 0, 1 ],[ 1, 0 ],[ -1, 1 ]]]),
(-8, 2, [[[ -1, 0 ],[ 1, 1 ],[ -1, 2 ],[ 4, -1 ]],[[ -1, 1 ],[ 1, 0 ],[ 0, 2 ],[ 1, -1 ]],[[ -1, 0 ],[ 0, 0 ],[ -1, 1 ],[ 1, 0 ]],[[ -1, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]]]),
(-3, 3, [[[ 1, 0 ],[ -1//2, 1//2 ],[ 1//2, -1//2 ],[ -1//2, 1//2 ]],[[ -1//2, 1//2 ],[ -1//2, -1//2 ],[ 0, 0 ],[ -1//2, -1//2 ]],[[ 1//2, -1//2 ],[ -1//2, 1//2 ],[ 1//2, -1//2 ],[ 0, 0 ]]]),
(-20, 5, [[[ 0, 1 ],[ -2, 0 ],[ -2, 0 ],[ 0, -1 ]],[[ 1, 0 ],[ 0, 0 ],[ 0, 1 ],[ -1, 0 ]],[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ 0, 2 ],[ -4, -1 ],[ -4, 1 ],[ 0, -2 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ 2, 0 ],[ 0, 1 ],[ 0, 1 ],[ -2, 0 ]]]),
(-24, 6, [[[ -2, -2 ],[ 3, 0 ],[ 3, -2 ],[ 2, 1 ]],[[ 1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ -3, 4 ],[ -7, -3 ],[ -10, -1 ],[ 7, -3 ]],[[ -11, -2 ],[ 11, -2 ],[ 0, -6 ],[ 11, 4 ]],[[ -2, 1 ],[ -1, -1 ],[ -3, 0 ],[ 1, -1 ]]]),
(-7, 7, [[[ 1, 0 ],[ -1, 0 ],[ 0, 0 ],[ -1, 0 ]],[[ 1, 0 ],[ -1//2, 1//2 ],[ 0, 0 ],[ -1, 0 ]],[[ -1, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -1//2, -1//2 ],[ 1, 0 ],[ 1//2, -1//2 ],[ 1//2, 1//2 ]]]),
(-40, 10, [[[ 9, -2 ],[ 0, 4 ],[ 9, 1 ],[ -9, 2 ]],[[ -2, 1 ],[ -3, -1 ],[ -3, 0 ],[ 1, -1 ]],[[ -1, 0 ],[ 2, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -11, -2 ],[ 11, -2 ],[ 0, -4 ],[ 11, 2 ]],[[ 3, 1 ],[ -6, 0 ],[ 0, 1 ],[ -3, -1 ]]]),
(-11, 11, [[[ 5//2, 1//2 ],[ 0, -1 ],[ 2, 0 ],[ -3//2, -1//2 ]],[[ -1, 0 ],[ 1//2, -1//2 ],[ -1//2, 1//2 ],[ -3//2, -1//2 ]],[[ -3//2, 1//2 ],[ 2, 0 ],[ 1//2, 1//2 ],[ 1//2, -1//2 ]]]),
(-52, 13, [[[ -4, 0 ],[ 2, -1 ],[ -2, -1 ],[ 4, 0 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -12, -2 ],[ 12, -3 ],[ 0, -3 ],[ 12, 1 ]],[[ 4, -4 ],[ 13, 2 ],[ 13, 2 ],[ -10, 3 ]],[[ -4, -3 ],[ 11, -1 ],[ 7, -2 ],[ 7, 2 ]],[[ 0, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ 0, -1 ],[ 3, 0 ],[ 4, 0 ],[ 0, 1 ]]]),
(-56, 14, [[[ -8, 0 ],[ 3, -2 ],[ -3, -2 ],[ 8, 0 ]],[[ 12, -2 ],[ 1, 2 ],[ 11, 2 ],[ -6, 1 ]],[[ 7, 1 ],[ -4, 1 ],[ -2, 3 ],[ -7, -1 ]],[[ 1, -1 ],[ 4, 0 ],[ 4, 0 ],[ 1, 1 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 1 ],[ -1, 0 ]],[[ 0, 4 ],[ -13, -2 ],[ -13, 2 ],[ 0, -4 ]]]),
(-15, 15, [[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ 1//2, -1//2 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 7//2, 1//2 ],[ 0, -1 ],[ 7//2, -1//2 ],[ -7//2, -1//2 ]],[[ 0, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]]]),
(-68, 17, [[[ -8, -5 ],[ 20, 0 ],[ 18, -4 ],[ 8, 5 ]],[[ 5, 0 ],[ -1, 1 ],[ 2, 2 ],[ -7, 0 ]],[[ -11, 2 ],[ -4, -3 ],[ -12, -2 ],[ 11, -2 ]],[[ 24, -5 ],[ 12, 6 ],[ 30, 5 ],[ -24, 5 ]],[[ -5, 1 ],[ -3, -2 ],[ -8, -1 ],[ 9, -2 ]],[[ 11, -7 ],[ 21, 4 ],[ 35, 0 ],[ -9, 7 ]],[[ -19, 8 ],[ -19, -8 ],[ -31, 1 ],[ 12, -7 ]],[[ -1, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 5, 0 ],[ -2, 1 ],[ 3, -1 ],[ 2, 1 ]]]),
(-19, 19, [[[ 2, 0 ],[ -3//2, 1//2 ],[ 1//2, -1//2 ],[ 3//2, 1//2 ]],[[ -3//2, 1//2 ],[ -3//2, -1//2 ],[ -2, 0 ],[ 1//2, -1//2 ]],[[ -3, 0 ],[ 3//2, -1//2 ],[ -3, -1 ],[ 5, 0 ]],[[ -3//2, 1//2 ],[ -1//2, -1//2 ],[ -3, 0 ],[ 3//2, -1//2 ]]]),
(-84, 21, [[[ 13, 1 ],[ -12, 3 ],[ -2, 3 ],[ -16, -2 ]],[[ -8, -1 ],[ 8, -1 ],[ 0, -2 ],[ 8, 1 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ 20, 3 ],[ -14, 3 ],[ -12, 6 ],[ -20, -3 ]],[[ 0, -1 ],[ 5, 0 ],[ 4, 0 ],[ 0, 1 ]],[[ -13, -2 ],[ 7, -2 ],[ 12, -4 ],[ 13, 2 ]],[[ 2, -1 ],[ 4, 1 ],[ 4, 0 ],[ -2, 1 ]],[[ 5, 8 ],[ -40, 0 ],[ -33, 2 ],[ -5, -8 ]],[[ 5, 0 ],[ 0, 1 ],[ 0, 1 ],[ -4, 0 ]],[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ -26, -5 ],[ 30, -5 ],[ 16, -6 ],[ 26, 5 ]]]),
(-88, 22, [[[ -13, 3 ],[ -3, -3 ],[ -23, 0 ],[ 10, -3 ]],[[ -8, 2 ],[ -1, -2 ],[ -17, 0 ],[ 9, -2 ]],[[ 5, -1 ],[ 1, 1 ],[ 8, 0 ],[ -3, 1 ]],[[ -25, 3 ],[ 6, -6 ],[ -27, -2 ],[ 25, -3 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 1, 0 ],[ -1, 0 ],[ 2, 0 ],[ -1, 0 ]],[[ -19, 5 ],[ -5, -5 ],[ -38, 0 ],[ 19, -5 ]]]),
(-23, 23, [[[ 3//2, 1//2 ],[ 3//2, -1//2 ],[ 3, 0 ],[ -3//2, -1//2 ]],[[ 1, 0 ],[ 1//2, -1//2 ],[ 1, 0 ],[ -1//2, -1//2 ]],[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 5//2, 1//2 ],[ 5//2, -1//2 ],[ 5//2, -1//2 ],[ -5//2, -1//2 ]],[[ 0, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]]]),
(-104, 26, [[[ 5, 0 ],[ 1, 1 ],[ 3, 2 ],[ -10, 1 ]],[[ 7, -1 ],[ 7, 1 ],[ 9, 1 ],[ -2, 2 ]],[[ 7, 0 ],[ -1, 1 ],[ 1, 1 ],[ -4, 0 ]],[[ 3, -1 ],[ 3, 1 ],[ 6, 0 ],[ -3, 1 ]],[[ 25, -3 ],[ 18, 6 ],[ 13, 4 ],[ -25, 3 ]],[[ -7, -2 ],[ 9, -2 ],[ 6, -1 ],[ 7, 1 ]],[[ 23, 1 ],[ -9, 5 ],[ -2, 4 ],[ -23, -1 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ -26, 0 ],[ -5, -5 ],[ 5, -5 ],[ 26, 0 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]]]),
(-116, 29, [[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ -28, -3 ],[ 28, -3 ],[ 0, -6 ],[ 28, 3 ]],[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ -11, -2 ],[ 16, -1 ],[ 11, -2 ],[ 5, 3 ]],[[ -34, 0 ],[ 10, -5 ],[ -14, -7 ],[ 34, 0 ]],[[ 2, 1 ],[ -4, 0 ],[ -6, 1 ],[ -2, -1 ]],[[ 3, 3 ],[ -18, -1 ],[ -21, 2 ],[ 0, -5 ]],[[ 21, -2 ],[ 2, 4 ],[ 17, 2 ],[ -15, 2 ]],[[ -20, -2 ],[ 15, -2 ],[ -1, -6 ],[ 24, 2 ]],[[ -5, 2 ],[ -9, -2 ],[ -10, 0 ],[ 5, -2 ]],[[ -18, 4 ],[ -11, -4 ],[ -25, -2 ],[ 17, -3 ]],[[ -13, 3 ],[ -7, -2 ],[ -20, -1 ],[ 7, -2 ]]]),
(-120, 30, [[[ 3, 1 ],[ -5, 0 ],[ -5, 1 ],[ -2, -1 ]],[[ 233, 83 ],[ -466, 0 ],[ -327, 83 ],[ -233, -83 ]],[[ 41, -4 ],[ 0, 8 ],[ 41, 5 ],[ -41, 4 ]],[[ 11, -1 ],[ 5, 2 ],[ 6, 2 ],[ -11, 1 ]],[[ 29, -3 ],[ 6, 6 ],[ 25, 4 ],[ -29, 3 ]],[[ 4, -1 ],[ 3, 1 ],[ 9, 1 ],[ -8, 1 ]],[[ 0, 4 ],[ -19, -2 ],[ -19, 2 ],[ 0, -4 ]],[[ 50, 18 ],[ -101, 0 ],[ -71, 18 ],[ -51, -18 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -97, 9 ],[ 0, -18 ],[ -100, -14 ],[ 103, -9 ]],[[ -23, 1 ],[ 4, -4 ],[ -16, -4 ],[ 25, -1 ]],[[ 5, 0 ],[ -1, 1 ],[ 1, 1 ],[ -6, 0 ]]]),
(-31, 31, [[[ -1//2, -1//2 ],[ 3, 0 ],[ 3, 0 ],[ -1//2, 1//2 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ -1//2, -1//2 ]],[[ 11//2, -1//2 ],[ 1//2, 3//2 ],[ 7//2, 1//2 ],[ -11//2, 1//2 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 3//2, -1//2 ],[ 3//2, 1//2 ],[ 3, 0 ],[ -3//2, 1//2 ]],[[ 3, 0 ],[ -3//2, 1//2 ],[ 3//2, 1//2 ],[ -3, 0 ]]]),
(-132, 33, [[[ 11, 0 ],[ -1, 2 ],[ 1, 2 ],[ -12, 0 ]],[[ 34, 0 ],[ 0, 5 ],[ 0, 7 ],[ -34, 0 ]],[[ -10, 0 ],[ 0, -1 ],[ 0, -3 ],[ 10, 0 ]],[[ 20, 1 ],[ -7, 3 ],[ -4, 4 ],[ -20, -1 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 0, 1 ],[ -4, 0 ],[ -8, 0 ],[ 0, -1 ]],[[ -55, 2 ],[ -2, -9 ],[ -17, -11 ],[ 60, -1 ]],[[ -33, 7 ],[ -31, -6 ],[ -50, -5 ],[ 32, -7 ]],[[ 0, 2 ],[ -10, -1 ],[ -10, 1 ],[ 0, -2 ]],[[ 1, 1 ],[ -5, 0 ],[ -7, 0 ],[ 1, -1 ]],[[ 0, 4 ],[ -23, 0 ],[ -23, 0 ],[ 0, -4 ]],[[ 7, 1 ],[ -6, 1 ],[ -5, 2 ],[ -10, -1 ]],[[ -32, 3 ],[ -12, -6 ],[ -22, -5 ],[ 32, -3 ]],[[ 100, 1 ],[ -20, 15 ],[ 0, 20 ],[ -100, -3 ]]]),
(-136, 34, [[[ 31, -3 ],[ 6, 6 ],[ 27, 4 ],[ -31, 3 ]],[[ 5, 2 ],[ -9, 0 ],[ -13, 2 ],[ -4, -2 ]],[[ -17, 5 ],[ -11, -6 ],[ -31, 1 ],[ 18, -5 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 13, -6 ],[ 32, 4 ],[ 31, -2 ],[ 3, 6 ]],[[ 5, 2 ],[ -10, 0 ],[ -11, 2 ],[ -5, -2 ]],[[ 29, -5 ],[ 15, 5 ],[ 46, 4 ],[ -29, 5 ]],[[ 11, 3 ],[ -16, 1 ],[ -18, 3 ],[ -11, -3 ]],[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ -1, -1 ],[ 6, 0 ],[ 6, 0 ],[ -1, 1 ]],[[ 7, 4 ],[ -26, 0 ],[ -29, 3 ],[ -9, -6 ]]]),
(-35, 35, [[[ -3, -1 ],[ 13//2, -1//2 ],[ 11//2, -1//2 ],[ 3, 1 ]],[[ 9//2, -1//2 ],[ 1//2, 1//2 ],[ 3//2, 1//2 ],[ -2, 0 ]],[[ 5//2, -1//2 ],[ 3//2, 1//2 ],[ 4, 0 ],[ -3//2, 1//2 ]],[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]]]),
(-148, 37, [[[ -7, 0 ],[ 2, -1 ],[ -2, -1 ],[ 6, 0 ]],[[ 25, -6 ],[ 25, 6 ],[ 48, 1 ],[ -23, 7 ]],[[ 19, -5 ],[ 14, 5 ],[ 39, 0 ],[ -20, 5 ]],[[ -36, 0 ],[ 0, -5 ],[ 0, -7 ],[ 36, 0 ]],[[ -7, 2 ],[ -7, -2 ],[ -14, 0 ],[ 7, -2 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -31, 2 ],[ 4, -5 ],[ -21, -4 ],[ 29, -1 ]],[[ 84, 1 ],[ -42, 11 ],[ 42, 13 ],[ -84, 0 ]],[[ -32, 0 ],[ 10, -5 ],[ -10, -5 ],[ 32, 0 ]],[[ 34, 3 ],[ -24, 5 ],[ -12, 6 ],[ -34, -3 ]]]),
(-152, 38, [[[ 1, 1 ],[ -5, 0 ],[ -7, 1 ],[ -4, -1 ]],[[ 18, 1 ],[ -15, 2 ],[ 1, 4 ],[ -22, -2 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ 3, 2 ],[ -13, 0 ],[ -17, 1 ],[ -2, -3 ]],[[ 39, 3 ],[ -39, 4 ],[ 0, 7 ],[ -39, -4 ]],[[ -30, -2 ],[ 27, -3 ],[ 9, -5 ],[ 21, 4 ]],[[ -11, 3 ],[ -10, -3 ],[ -22, 0 ],[ 11, -3 ]],[[ -26, 6 ],[ -17, -7 ],[ -41, -3 ],[ 39, -4 ]],[[ 37, -3 ],[ 18, 6 ],[ 19, 6 ],[ -37, 3 ]],[[ 9, -2 ],[ 5, 1 ],[ 14, 1 ],[ -5, 1 ]],[[ -11, -1 ],[ 6, -1 ],[ 6, -2 ],[ 7, 1 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 45, -7 ],[ 14, 10 ],[ 54, 5 ],[ -55, 5 ]],[[ 11, 3 ],[ -21, 0 ],[ -11, 3 ],[ -10, -3 ]]]),
(-39, 39, [[[ -5, 0 ],[ 11//2, -3//2 ],[ -7//2, -1//2 ],[ 19//2, -1//2 ]],[[ -5//2, -1//2 ],[ 5, 0 ],[ 1//2, -1//2 ],[ 5//2, 1//2 ]],[[ -11//2, 1//2 ],[ 0, -1 ],[ -11//2, -1//2 ],[ 11//2, -1//2 ]],[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ 3//2, -1//2 ],[ -1//2, 1//2 ],[ 1, 0 ],[ -1, 0 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ 7//2, 1//2 ],[ -7, 0 ],[ 1//2, 1//2 ],[ -7//2, -1//2 ]],[[ 2, 1 ],[ -21//2, -1//2 ],[ -5//2, 1//2 ],[ -2, -1 ]]]),
(-164, 41, [[[ -61, 2 ],[ 25, -9 ],[ -35, -6 ],[ 52, -1 ]],[[ -7, -2 ],[ 17, 0 ],[ 18, -2 ],[ 3, 4 ]],[[ 128, 3 ],[ -109, 17 ],[ 51, 15 ],[ -128, -3 ]],[[ 3, 1 ],[ -7, 0 ],[ -4, 1 ],[ -4, -1 ]],[[ -81, 25 ],[ -84, -29 ],[ -154, 1 ],[ 104, -22 ]],[[ 26, -5 ],[ 18, 7 ],[ 33, 4 ],[ -45, 3 ]],[[ 48, -3 ],[ -3, 7 ],[ 45, 4 ],[ -37, 4 ]],[[ -74, 11 ],[ -29, -18 ],[ -84, -4 ],[ 74, -11 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -13, -8 ],[ 44, 0 ],[ 57, -4 ],[ 9, 8 ]],[[ 8, 13 ],[ -89, -6 ],[ -51, 7 ],[ -17, -12 ]],[[ -48, -19 ],[ 104, -2 ],[ 132, -15 ],[ 48, 19 ]],[[ 25, -14 ],[ 64, 6 ],[ 74, -5 ],[ 9, 10 ]],[[ -12, -7 ],[ 39, 0 ],[ 23, -5 ],[ 21, 4 ]],[[ 56, -3 ],[ -23, 10 ],[ 37, 10 ],[ -84, -2 ]]]),
(-168, 42, [[[ -55, -6 ],[ 55, -3 ],[ 0, -12 ],[ 55, 6 ]],[[ -7, 2 ],[ -5, -2 ],[ -18, 0 ],[ 11, -2 ]],[[ 449, 40 ],[ -449, 40 ],[ 0, 80 ],[ -449, -40 ]],[[ -11, -1 ],[ 11, -1 ],[ 4, -3 ],[ 15, 2 ]],[[ 97, 0 ],[ -51, 13 ],[ 51, 13 ],[ -100, 0 ]],[[ 23, 0 ],[ -14, 3 ],[ 14, 3 ],[ -25, 0 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ 1, 6 ],[ -34, -3 ],[ -37, 4 ],[ -3, -7 ]],[[ 37, 5 ],[ -51, 2 ],[ -4, 9 ],[ -49, -6 ]],[[ 28, 72 ],[ -429, -42 ],[ -333, 42 ],[ -28, -72 ]],[[ -13, -2 ],[ 13, 0 ],[ 0, -4 ],[ 13, 2 ]],[[ -29, 7 ],[ -21, -7 ],[ -58, 0 ],[ 29, -7 ]]]),
(-43, 43, [[[ -3//2, -1//2 ],[ 4, 0 ],[ 3//2, -1//2 ],[ 5//2, 1//2 ]],[[ -5//2, 1//2 ],[ -1, -1 ],[ -3, 0 ],[ 7//2, -1//2 ]],[[ -3//2, -1//2 ],[ 4, 0 ],[ -3, 0 ],[ 3//2, -1//2 ]],[[ -2, 4 ],[ -49//2, -7//2 ],[ -37//2, 3//2 ],[ 5, -4 ]]]),
(-184, 46, [[[ 133, 6 ],[ -160, 14 ],[ 35, 19 ],[ -159, -12 ]],[[ -31, 14 ],[ -71, -16 ],[ -72, 4 ],[ 31, -14 ]],[[ 41, 7 ],[ -82, 0 ],[ -7, 7 ],[ -41, -7 ]],[[ 61, 2 ],[ -48, 7 ],[ 20, 8 ],[ -61, -2 ]],[[ 0, 8 ],[ -47, -4 ],[ -47, 4 ],[ 0, -8 ]],[[ 11, 0 ],[ -10, 2 ],[ 5, 1 ],[ -13, 0 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ -91, 6 ],[ 35, -19 ],[ -67, -5 ],[ 92, -6 ]],[[ 64, -7 ],[ 5, 12 ],[ 73, 4 ],[ -64, 7 ]],[[ -7, -4 ],[ 33, 2 ],[ 18, -3 ],[ 7, 5 ]],[[ -87, -8 ],[ 95, -7 ],[ -15, -11 ],[ 74, 4 ]],[[ 95, 4 ],[ -77, 10 ],[ 28, 10 ],[ -75, -2 ]],[[ -50, -8 ],[ 99, 0 ],[ 5, -8 ],[ 49, 8 ]]]),
(-47, 47, [[[ -10, -1 ],[ 10, -1 ],[ 2, -2 ],[ 12, 1 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ 7//2, -1//2 ],[ 7//2, 1//2 ],[ 7//2, 1//2 ],[ -7//2, 1//2 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 5//2, 1//2 ],[ -9//2, 1//2 ],[ -5//2, 1//2 ],[ -9//2, -1//2 ]],[[ 4, 1 ],[ -8, 0 ],[ -4, 1 ],[ -4, -1 ]],[[ 3, 0 ],[ -3//2, 1//2 ],[ 3//2, 1//2 ],[ -5, 0 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]]]),
(-51, 51, [[[ 8, 4 ],[ 41//2, -11//2 ],[ 33//2, 3//2 ],[ -7, -4 ]],[[ 23, -3 ],[ -47, 0 ],[ 2, -3 ],[ -24, 3 ]],[[ 119//2, -5//2 ],[ -82, -5 ],[ 23, -5 ],[ -117//2, 5//2 ]],[[ -21//2, 1//2 ],[ 8, 1 ],[ -4, 1 ],[ 8, 0 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 148, -37 ],[ -414, 22 ],[ -77//2, -57//2 ],[ -148, 37 ]],[[ 0, -1 ],[ -7, 1 ],[ -7//2, -1//2 ],[ 0, 1 ]]]),
(-212, 53, [[[ -1, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -12, 1 ],[ 17, 2 ],[ 0, 2 ],[ 22, -1 ]],[[ 67, -1 ],[ -41, -8 ],[ 28, -5 ],[ -48, -1 ]],[[ -160, -43 ],[ -212, 42 ],[ -320, 0 ],[ 160, 41 ]],[[ 937, 36 ],[ -216, -144 ],[ 602, -89 ],[ -937, -36 ]],[[ 54, 5 ],[ 0, -10 ],[ 54, -3 ],[ -54, -5 ]],[[ -14, 3 ],[ 28, 0 ],[ 10, 3 ],[ 14, -3 ]],[[ -18, 56 ],[ 358, -25 ],[ 322, 25 ],[ 0, -50 ]],[[ -21, 1 ],[ 25, 2 ],[ -4, 3 ],[ 25, -2 ]],[[ -15, 10 ],[ 79, -3 ],[ 37, 4 ],[ 10, -7 ]],[[ -78, -11 ],[ 0, 22 ],[ -110, 4 ],[ 142, 11 ]],[[ -21, 8 ],[ 63, -2 ],[ 31, 4 ],[ 8, -6 ]],[[ -1, 0 ],[ 2, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ 0, -22 ],[ -148, 11 ],[ -130, -11 ],[ -9, 22 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 4, 1 ],[ 6, -1 ],[ 7, 0 ],[ -3, -1 ]]]),
(-55, 55, [[[ 13//2, -1//2 ],[ 0, 1 ],[ 13//2, 1//2 ],[ -13//2, 1//2 ]],[[ 15//2, -1//2 ],[ -7//2, 3//2 ],[ 7//2, 1//2 ],[ -7, 0 ]],[[ 9//2, 1//2 ],[ -7, 0 ],[ -1//2, 1//2 ],[ -5//2, -1//2 ]],[[ -3//2, 1//2 ],[ -7//2, -1//2 ],[ -3, 0 ],[ 3//2, -1//2 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ 0, 1 ],[ -12, -1 ],[ -9//2, 1//2 ],[ -3//2, -3//2 ]],[[ 0, -1 ],[ 31//2, 1//2 ],[ 9//2, -1//2 ],[ 11//2, 3//2 ]]]),
(-228, 57, [[[ 757, -11 ],[ -384, -81 ],[ 251, -100 ],[ -748, 13 ]],[[ 155, 8 ],[ -12, -21 ],[ 128, -16 ],[ -157, -8 ]],[[ 352, 3 ],[ -141, -43 ],[ 175, -43 ],[ -370, -1 ]],[[ 44, 3 ],[ 3, -6 ],[ 40, -4 ],[ -40, -3 ]],[[ -13115, 1739 ],[ 17666, 751 ],[ 5704, 2339 ],[ 13102, -1740 ]],[[ 355, 18 ],[ -41, -50 ],[ 283, -38 ],[ -384, -16 ]],[[ -9, 0 ],[ 5, 1 ],[ -5, 1 ],[ 9, 0 ]],[[ -247, 20 ],[ 231, 20 ],[ 32, 40 ],[ 247, -20 ]],[[ -17913, 638 ],[ 12469, 1819 ],[ -3899, 2404 ],[ 17926, -639 ]],[[ 313, 22 ],[ 4, -47 ],[ 288, -26 ],[ -307, -22 ]],[[ 53, 4 ],[ 8, -8 ],[ 48, -5 ],[ -53, -4 ]],[[ 79, 7 ],[ 15, -12 ],[ 84, -7 ],[ -80, -7 ]],[[ -85, -50 ],[ -294, 31 ],[ -388, -12 ],[ 85, 50 ]],[[ 21, -8 ],[ -59, 1 ],[ -50, -6 ],[ -17, 8 ]],[[ -107, -5 ],[ 11, 14 ],[ -84, 11 ],[ 104, 5 ]],[[ -105, -8 ],[ -4, 17 ],[ -109, 9 ],[ 118, 9 ]],[[ -91, -6 ],[ -2, 13 ],[ -84, 9 ],[ 94, 6 ]]]),
(-232, 58, [[[ 37, 0 ],[ -12, 4 ],[ 12, 4 ],[ -29, 0 ]],[[ 16, 0 ],[ -5, 2 ],[ 5, 2 ],[ -16, 0 ]],[[ -1, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -1, -2 ],[ 16, 0 ],[ 14, -1 ],[ 7, 2 ]],[[ 44, 2 ],[ -35, 4 ],[ -9, 6 ],[ -35, -4 ]],[[ 37, 9 ],[ -74, 0 ],[ -45, 9 ],[ -37, -9 ]],[[ -345, 8 ],[ 73, -45 ],[ -226, -36 ],[ 339, -14 ]],[[ 15, 2 ],[ -23, 1 ],[ -15, 2 ],[ -8, -3 ]],[[ 39, 2 ],[ -29, 4 ],[ -9, 5 ],[ -32, -3 ]],[[ -1451, 110 ],[ 0, -220 ],[ -1451, -110 ],[ 1451, -110 ]],[[ -1, 5 ],[ -35, -2 ],[ -43, 2 ],[ 2, -6 ]],[[ 22, 0 ],[ -7, 3 ],[ 7, 3 ],[ -26, 0 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 43, 0 ],[ -21, 5 ],[ 21, 5 ],[ -44, 0 ]]]),
(-59, 59, [[[ 1, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ 2, 0 ],[ 1//2, 1//2 ],[ 3//2, 1//2 ],[ -15//2, 1//2 ]],[[ 23//2, -1//2 ],[ 0, 1 ],[ 9, 1 ],[ -13//2, 1//2 ]],[[ 0, -1 ],[ 12, 0 ],[ 5, 0 ],[ 0, 1 ]],[[ -5//2, -1//2 ],[ 5, 0 ],[ -4, 0 ],[ 5//2, -1//2 ]],[[ -1//2, -1//2 ],[ 4, 0 ],[ 4, 0 ],[ -1//2, 1//2 ]]]),
(-244, 61, [[[ -10, -3 ],[ 25, 0 ],[ 16, -3 ],[ 15, 3 ]],[[ 8, 1 ],[ -9, 1 ],[ -6, 2 ],[ -16, -1 ]],[[ -62, 0 ],[ 0, -9 ],[ 0, -7 ],[ 62, 0 ]],[[ 59, 4 ],[ -38, 7 ],[ -24, 8 ],[ -59, -4 ]],[[ -31, 4 ],[ -31, -4 ],[ -47, -2 ],[ 16, -6 ]],[[ -92, -15 ],[ 120, -11 ],[ 49, -18 ],[ 137, 7 ]],[[ 86, 18 ],[ -173, 0 ],[ -71, 18 ],[ -87, -18 ]],[[ 18, 0 ],[ -9, 2 ],[ 9, 2 ],[ -18, 0 ]],[[ 60, 11 ],[ -88, 7 ],[ -42, 13 ],[ -98, -6 ]],[[ -10, -1 ],[ 9, -1 ],[ 7, -1 ],[ 6, 1 ]],[[ -40, 0 ],[ 2, -5 ],[ -18, -5 ],[ 39, -2 ]],[[ 84, 0 ],[ -46, 9 ],[ 46, 9 ],[ -84, 0 ]],[[ -88, -12 ],[ 107, -12 ],[ 31, -16 ],[ 137, 5 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ 5, 1 ],[ -9, 0 ],[ -9, 2 ],[ -8, -2 ]],[[ -59, -11 ],[ 99, 0 ],[ 40, -13 ],[ 58, 11 ]]]),
(-248, 62, [[[ 319, 10 ],[ -230, 30 ],[ 89, 39 ],[ -311, -10 ]],[[ 603, -8 ],[ -250, 71 ],[ 350, 61 ],[ -597, 8 ]],[[ 69, -65 ],[ 412, 41 ],[ 470, -25 ],[ -69, 65 ]],[[ -63, -4 ],[ 63, -6 ],[ 0, -8 ],[ 63, 4 ]],[[ 78, -9 ],[ 33, 13 ],[ 99, 4 ],[ -71, 10 ]],[[ 8, 0 ],[ -1, 1 ],[ 1, 1 ],[ -8, 0 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ 113, -25 ],[ 117, 25 ],[ 224, 0 ],[ -111, 25 ]],[[ 41, -3 ],[ 9, 8 ],[ 41, 3 ],[ -50, 5 ]],[[ 67, 42 ],[ -327, -14 ],[ -246, 28 ],[ -69, -42 ]],[[ 331, -115 ],[ 626, 95 ],[ 929, -21 ],[ -325, 114 ]],[[ 203, -15 ],[ 0, 30 ],[ 200, 14 ],[ -197, 15 ]],[[ 377, 81 ],[ -754, 0 ],[ -351, 81 ],[ -377, -81 ]],[[ 181, 39 ],[ -362, 0 ],[ -170, 39 ],[ -181, -39 ]],[[ -23, 3 ],[ -19, -2 ],[ -31, -1 ],[ 4, -3 ]],[[ -11, -2 ],[ 13, -1 ],[ 9, -2 ],[ 12, 1 ]],[[ 301, 7 ],[ -200, 30 ],[ 100, 36 ],[ -299, -7 ]],[[ 116, -28 ],[ 133, 27 ],[ 249, 1 ],[ -133, 27 ]],[[ 5, -1 ],[ 6, 1 ],[ 8, 0 ],[ -3, 1 ]]]),
(-260, 65, [[[ 4, -6 ],[ -59, 2 ],[ -35, -2 ],[ -4, 6 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -61, 4 ],[ 40, 7 ],[ 24, 8 ],[ 61, -4 ]],[[ 7, 0 ],[ -2, -1 ],[ 2, -1 ],[ -10, 0 ]],[[ -14, 1 ],[ 14, 1 ],[ 0, 2 ],[ 14, -1 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ -64, 0 ],[ 0, 9 ],[ 0, 7 ],[ 64, 0 ]],[[ -2, 6 ],[ 41, 0 ],[ 47, 2 ],[ 12, -5 ]],[[ 1, -3 ],[ -26, 1 ],[ -20, -1 ],[ -1, 3 ]],[[ 206, 19 ],[ 108, -23 ],[ 201, -29 ],[ -213, -18 ]],[[ -98, 15 ],[ 178, 7 ],[ 64, 14 ],[ 98, -15 ]],[[ -46, -3 ],[ -16, 5 ],[ -22, 3 ],[ 22, 2 ]],[[ 92, 9 ],[ 55, -10 ],[ 61, -5 ],[ -37, -6 ]],[[ 224, 7 ],[ 40, -24 ],[ 96, -32 ],[ -224, -9 ]],[[ 6, 0 ],[ 0, -1 ],[ 0, -1 ],[ -11, 0 ]],[[ -1, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 26, 0 ],[ -8, -3 ],[ 8, -3 ],[ -25, 0 ]],[[ 0, 10 ],[ 79, -2 ],[ 79, 2 ],[ 0, -10 ]],[[ 10, 1 ],[ 6, -2 ],[ 9, -1 ],[ -14, -1 ]],[[ -4, -1 ],[ -6, 1 ],[ -8, 0 ],[ 4, 1 ]],[[ -33, 0 ],[ 9, 4 ],[ -9, 4 ],[ 34, 0 ]],[[ 2, 1 ],[ 12, -1 ],[ 5, 0 ],[ -3, -1 ]]]),
(-264, 66, [[[ 15, -1 ],[ -2, 2 ],[ 22, 2 ],[ -25, 1 ]],[[ 20, -4 ],[ 17, 3 ],[ 49, 1 ],[ -21, 4 ]],[[ 273, -14 ],[ -4, 28 ],[ 236, 21 ],[ -193, 14 ]],[[ -11, 5 ],[ -25, -4 ],[ -40, 2 ],[ 11, -5 ]],[[ 423, 54 ],[ -436, 21 ],[ -313, 87 ],[ -409, -53 ]],[[ -32, 5 ],[ -17, -6 ],[ -49, -2 ],[ 40, -4 ]],[[ 75, -4 ],[ -4, 9 ],[ 71, 5 ],[ -61, 5 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -407, -8 ],[ 272, -32 ],[ -213, -49 ],[ 407, 8 ]],[[ -5, 33 ],[ -194, -19 ],[ -321, 16 ],[ 85, -38 ]],[[ 177, -13 ],[ -20, 24 ],[ 199, 10 ],[ -175, 13 ]],[[ 4, -1 ],[ 5, 1 ],[ 9, 0 ],[ -5, 1 ]],[[ -120, 9 ],[ -17, -20 ],[ -143, -7 ],[ 131, -15 ]],[[ -1312, -68 ],[ 1159, -81 ],[ -387, -181 ],[ 1312, 68 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ 75, -19 ],[ 124, 19 ],[ 148, 0 ],[ -73, 19 ]],[[ -77, 3 ],[ 14, -10 ],[ -70, -6 ],[ 77, -5 ]],[[ -200, -10 ],[ 127, -15 ],[ -7, -35 ],[ 214, 11 ]],[[ -16, -8 ],[ 57, 3 ],[ 45, -7 ],[ 17, 8 ]],[[ -390, -23 ],[ 251, -27 ],[ 23, -69 ],[ 390, 23 ]],[[ 103, -44 ],[ 225, 22 ],[ 372, -18 ],[ -5, 38 ]],[[ -11, 14 ],[ -78, -9 ],[ -177, 5 ],[ 71, -19 ]]]),
(-67, 67, [[[ 11//2, -3//2 ],[ -13, 0 ],[ -17//2, -3//2 ],[ -15//2, 3//2 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -5//2, -1//2 ],[ -3//2, 1//2 ],[ -6, 0 ],[ 7//2, 1//2 ]],[[ 0, -7 ],[ -99//2, 7//2 ],[ -109//2, -7//2 ],[ 5//2, 15//2 ]]]),
(-276, 69, [[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ -15, -4 ],[ 35, 0 ],[ 23, -4 ],[ 20, 4 ]],[[ 58, 13 ],[ -145, 1 ],[ -58, 13 ],[ -87, -14 ]],[[ -142, -89 ],[ 708, 35 ],[ 532, -62 ],[ 142, 89 ]],[[ -56, 15 ],[ -92, -11 ],[ -134, 2 ],[ 30, -15 ]],[[ -46, -15 ],[ 122, 3 ],[ 79, -13 ],[ 47, 14 ]],[[ -57, -23 ],[ 181, 6 ],[ 125, -18 ],[ 57, 21 ]],[[ 34, 3 ],[ -44, 3 ],[ -23, 4 ],[ -25, -5 ]],[[ -49, 13 ],[ -65, -12 ],[ -117, 2 ],[ 49, -13 ]],[[ -11, 2 ],[ -7, -2 ],[ -22, 0 ],[ 11, -2 ]],[[ 22, 1 ],[ -22, 3 ],[ 0, 2 ],[ -22, -1 ]],[[ -50, 0 ],[ 26, -5 ],[ -26, -5 ],[ 48, 0 ]],[[ 12, -1 ],[ 5, 1 ],[ 13, 1 ],[ -6, 1 ]],[[ -5, 3 ],[ -24, -1 ],[ -24, 1 ],[ -5, -3 ]],[[ 3, -9 ],[ 62, 3 ],[ 80, -3 ],[ -3, 9 ]],[[ 0, 0 ],[ -1, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ -42, 1 ],[ 16, -5 ],[ -22, -6 ],[ 56, 1 ]],[[ -120, -9 ],[ 127, -8 ],[ -7, -17 ],[ 127, 8 ]],[[ 36, -25 ],[ 151, 17 ],[ 194, -9 ],[ -43, 24 ]],[[ -50, -2 ],[ 43, -4 ],[ 15, -6 ],[ 34, 5 ]],[[ -44, -3 ],[ 38, -4 ],[ 9, -6 ],[ 44, 3 ]]]),
(-280, 70, [[[ 53, 9 ],[ -106, 0 ],[ -27, 9 ],[ -53, -9 ]],[[ -31, 1 ],[ 8, -4 ],[ -18, -2 ],[ 25, -1 ]],[[ -43, 0 ],[ 30, -6 ],[ -20, -4 ],[ 53, 0 ]],[[ -35, -2 ],[ 44, -4 ],[ -4, -4 ],[ 45, 2 ]],[[ 107, -33 ],[ 199, 32 ],[ 258, -6 ],[ -107, 33 ]],[[ -5, -1 ],[ 12, 0 ],[ 3, -1 ],[ 7, 1 ]],[[ -8, -3 ],[ 33, 2 ],[ 17, -2 ],[ 0, 4 ]],[[ 31, -14 ],[ 95, 13 ],[ 96, -4 ],[ -31, 14 ]],[[ -48, -4 ],[ 63, -3 ],[ 7, -5 ],[ 36, 4 ]],[[ 33, 0 ],[ -20, 4 ],[ 10, 2 ],[ -23, 0 ]],[[ -41, 17 ],[ -117, -15 ],[ -123, 4 ],[ 40, -17 ]],[[ -7, -3 ],[ 27, 1 ],[ 13, -1 ],[ 0, 2 ]],[[ -30, -4 ],[ 49, -1 ],[ 11, -3 ],[ 17, 3 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ 7, 1 ],[ -15, 0 ],[ -1, 1 ],[ -8, -1 ]],[[ 64, 90 ],[ -765, -48 ],[ -525, 48 ],[ -64, -90 ]],[[ -4, -2 ],[ 19, 1 ],[ 11, -1 ],[ 0, 2 ]],[[ -29, -2 ],[ 29, -2 ],[ 0, -4 ],[ 29, 2 ]],[[ 5, -1 ],[ 11, 2 ],[ 9, 0 ],[ -8, 2 ]],[[ -24, -15 ],[ 137, 6 ],[ 83, -9 ],[ 25, 15 ]]]),
(-71, 71, [[[ -1, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ -7//2, -3//2 ],[ -7//2, 3//2 ],[ -19//2, -1//2 ],[ 6, 1 ]],[[ 27//2, 1//2 ],[ -7, -2 ],[ 7//2, -1//2 ],[ -7, 0 ]],[[ 1, 0 ],[ -2, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ 19//2, -1//2 ],[ -19//2, -1//2 ],[ 2, -1 ],[ -15//2, 1//2 ]],[[ -8, -1 ],[ -1, 2 ],[ -8, 0 ],[ 8, 1 ]],[[ -7//2, 3//2 ],[ 16, -1 ],[ 9//2, 1//2 ],[ 0, -1 ]],[[ 0, 7 ],[ 59, -7 ],[ 59//2, 7//2 ],[ 0, -7 ]],[[ 5//2, 1//2 ],[ 5//2, -1//2 ],[ 5, 0 ],[ -5//2, -1//2 ]]]),
(-292, 73, [[[ -30, 7 ],[ -37, -7 ],[ -52, 1 ],[ 21, -6 ]],[[ -316, -163 ],[ 1738, 142 ],[ 474, -98 ],[ 316, 163 ]],[[ 4, -3 ],[ 19, 4 ],[ 15, -2 ],[ -2, 4 ]],[[ 299, 4 ],[ -265, 31 ],[ 128, 24 ],[ -299, -4 ]],[[ 1, 0 ],[ -2, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ -546, 64 ],[ 53, -134 ],[ -601, 8 ],[ 591, -79 ]],[[ -72, -3 ],[ 80, -6 ],[ -31, -6 ],[ 74, 1 ]],[[ -20, -29 ],[ 237, 12 ],[ 202, -15 ],[ 19, 29 ]],[[ 6, 1 ],[ -12, 0 ],[ -3, 1 ],[ -6, -1 ]],[[ -53, -5 ],[ 103, 0 ],[ -8, -5 ],[ 50, 5 ]],[[ 22, -2 ],[ -7, 4 ],[ 21, 0 ],[ -20, 2 ]],[[ 0, 12 ],[ -95, -6 ],[ -83, 6 ],[ -6, -12 ]],[[ -9, -1 ],[ 17, 0 ],[ 0, -1 ],[ 8, 1 ]],[[ -28, -7 ],[ 72, 4 ],[ 38, -6 ],[ 4, 9 ]],[[ -616, 192 ],[ -962, -283 ],[ -1050, 63 ],[ 616, -192 ]],[[ -14, 5 ],[ -22, -7 ],[ -27, 2 ],[ 15, -5 ]],[[ 0, 10 ],[ -74, -5 ],[ -74, 5 ],[ 0, -10 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ 43, 0 ],[ -49, 5 ],[ 22, 4 ],[ -59, -2 ]]]),
(-296, 74, [[[ -7, 7 ],[ 60, -1 ],[ 60, 1 ],[ -7, -7 ]],[[ 10, 3 ],[ 25, -1 ],[ 27, 0 ],[ 1, -3 ]],[[ 3, -1 ],[ -4, 0 ],[ -18, -1 ],[ -1, 1 ]],[[ -30, 3 ],[ 31, 3 ],[ 17, 6 ],[ 49, -3 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ 209, -13 ],[ -143, -22 ],[ -78, -26 ],[ -209, 13 ]],[[ 23, 1 ],[ 7, -3 ],[ 27, -4 ],[ -40, -3 ]],[[ 0, 1 ],[ 5, 0 ],[ 15, 0 ],[ 0, -1 ]],[[ 0, 0 ],[ -1, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ -7, 1 ],[ 12, 1 ],[ 6, 1 ],[ 11, -1 ]],[[ -1, 2 ],[ 16, 0 ],[ 18, 1 ],[ 7, -2 ]],[[ -29, 0 ],[ 0, 3 ],[ 0, 3 ],[ 23, 0 ]],[[ -48, -2 ],[ -21, 7 ],[ -21, 5 ],[ 54, 3 ]],[[ 83, 14 ],[ 104, -12 ],[ 133, -7 ],[ -83, -14 ]],[[ -41, -4 ],[ -33, 5 ],[ -27, 4 ],[ 36, 3 ]],[[ 7, 1 ],[ 10, -1 ],[ 22, -1 ],[ -11, -3 ]],[[ -5, -1 ],[ -5, 1 ],[ -10, 0 ],[ 5, 1 ]],[[ -8, -4 ],[ -17, 1 ],[ -35, 1 ],[ 9, 2 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 82, 0 ],[ -27, -9 ],[ 27, -9 ],[ -82, 0 ]],[[ -17, 0 ],[ 3, 2 ],[ -3, 2 ],[ 18, 0 ]]]),
(-308, 77, [[[ -45, -15 ],[ 132, -5 ],[ 87, -10 ],[ 87, 10 ]],[[ 0, 10 ],[ -76, -5 ],[ -76, 5 ],[ 0, -10 ]],[[ 28, -1 ],[ 13, 3 ],[ 15, 2 ],[ -15, 2 ]],[[ -73, -8 ],[ 73, -8 ],[ 31, -12 ],[ 104, 4 ]],[[ 23, 0 ],[ -6, 2 ],[ 6, 2 ],[ -15, 0 ]],[[ -116, -13 ],[ 176, -7 ],[ 46, -17 ],[ 130, 14 ]],[[ -83, 24 ],[ -144, -24 ],[ -196, 5 ],[ 83, -24 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ 65, -12 ],[ 65, 12 ],[ 132, 1 ],[ -67, 13 ]],[[ -120, 37 ],[ -231, -36 ],[ -298, 8 ],[ 120, -37 ]],[[ -79, -2 ],[ 63, -7 ],[ -17, -8 ],[ 74, 3 ]],[[ -21, 5 ],[ -34, -5 ],[ -43, 0 ],[ 22, -5 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -43, -6 ],[ 74, -2 ],[ 31, -5 ],[ 29, 6 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ -351, -20 ],[ 351, -30 ],[ 0, -40 ],[ 351, 20 ]],[[ -48, -3 ],[ 55, -3 ],[ 1, -6 ],[ 47, 4 ]],[[ 274, 15 ],[ -274, 25 ],[ 0, 30 ],[ -274, -15 ]],[[ 50, 0 ],[ -24, 5 ],[ 24, 5 ],[ -50, 0 ]],[[ 9, 2 ],[ -20, 0 ],[ -15, 1 ],[ -1, -2 ]],[[ -124, 20 ],[ -115, -24 ],[ -201, -4 ],[ 134, -21 ]],[[ 4, -1 ],[ 7, 1 ],[ 17, 0 ],[ -9, 2 ]],[[ -95, -3 ],[ 20, -11 ],[ -20, -11 ],[ 95, -3 ]]]),
(-312, 78, [[[ -5, -2 ],[ 20, 1 ],[ 12, -1 ],[ -1, 2 ]],[[ 14, -3 ],[ 17, 3 ],[ 29, 0 ],[ -15, 3 ]],[[ 27, 5 ],[ -59, -1 ],[ -23, 5 ],[ -22, -6 ]],[[ 16, 12 ],[ -105, -8 ],[ -57, 8 ],[ -16, -12 ]],[[ 7, 1 ],[ -14, 0 ],[ -2, 1 ],[ -7, -1 ]],[[ -175, -5 ],[ 194, -14 ],[ -69, -14 ],[ 175, 5 ]],[[ 55, 18 ],[ -207, -9 ],[ -61, 13 ],[ -64, -18 ]],[[ -71, -9 ],[ 142, 0 ],[ 9, -9 ],[ 71, 9 ]],[[ 25, -2 ],[ 0, 4 ],[ 25, 1 ],[ -25, 2 ]],[[ 245, -16 ],[ 0, 32 ],[ 204, 8 ],[ -163, 16 ]],[[ 23, 3 ],[ -44, 0 ],[ -5, 3 ],[ -21, -3 ]],[[ 11, -2 ],[ 13, 2 ],[ 18, 0 ],[ -7, 2 ]],[[ 376, 0 ],[ -187, 37 ],[ 187, 37 ],[ -377, 0 ]],[[ 25, -5 ],[ 31, 5 ],[ 46, 0 ],[ -21, 5 ]],[[ -209, -15 ],[ 254, -16 ],[ 11, -28 ],[ 249, 17 ]],[[ -41, 9 ],[ -59, -9 ],[ -80, 0 ],[ 39, -9 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 255, -12 ],[ -35, 31 ],[ 188, 19 ],[ -239, 9 ]],[[ 462, 7 ],[ -469, 44 ],[ 217, 39 ],[ -523, -11 ]],[[ 209, -20 ],[ 0, 40 ],[ 209, 4 ],[ -209, 20 ]],[[ -63, -1 ],[ 58, -6 ],[ -16, -6 ],[ 63, 3 ]]]),
(-79, 79, [[[ 9//2, 1//2 ],[ 9//2, -1//2 ],[ 9//2, -1//2 ],[ -9//2, -1//2 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ -12, 0 ],[ 8, 1 ],[ -8, 1 ],[ 12, 0 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -3, -1 ],[ -2, 1 ],[ -9, 0 ],[ 15//2, 1//2 ]],[[ 15//2, -1//2 ],[ -7, 0 ],[ 4, -1 ],[ -13//2, 1//2 ]],[[ 17, -2 ],[ -20, -1 ],[ -1, -3 ],[ -22, 1 ]],[[ -5, 0 ],[ 5//2, 1//2 ],[ -5//2, 1//2 ],[ 5, 0 ]],[[ 33//2, 1//2 ],[ -17//2, -3//2 ],[ 15//2, -3//2 ],[ -27//2, 1//2 ]]]),
(-328, 82, [[[ -79, -5 ],[ 73, -6 ],[ 10, -10 ],[ 79, 5 ]],[[ 249, -13 ],[ -16, 32 ],[ 235, 18 ],[ -269, 15 ]],[[ -19, 4 ],[ -25, -4 ],[ -38, 0 ],[ 19, -4 ]],[[ 5, -1 ],[ 7, 1 ],[ 9, 0 ],[ -4, 1 ]],[[ -155, 15 ],[ -36, -21 ],[ -177, -8 ],[ 143, -12 ]],[[ 55, 7 ],[ -73, 4 ],[ -27, 9 ],[ -70, -5 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -81, 15 ],[ -67, -15 ],[ -169, 0 ],[ 88, -15 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ -15, -3 ],[ 33, -1 ],[ 21, -2 ],[ 14, 3 ]],[[ 11, -2 ],[ 9, 2 ],[ 17, 1 ],[ -16, 1 ]],[[ -143, 1 ],[ 64, -14 ],[ -85, -14 ],[ 151, -1 ]],[[ -80, -5 ],[ 77, -5 ],[ 23, -10 ],[ 65, 7 ]],[[ 67, -13 ],[ 65, 13 ],[ 139, 0 ],[ -72, 13 ]],[[ -329, -22 ],[ 329, -19 ],[ 0, -44 ],[ 329, 22 ]],[[ -233, -15 ],[ 230, -14 ],[ 0, -30 ],[ 227, 15 ]],[[ 233, -99 ],[ 667, 72 ],[ 891, -28 ],[ -224, 100 ]],[[ -76, 0 ],[ 35, -7 ],[ -35, -7 ],[ 69, 0 ]],[[ -1162, 0 ],[ 587, -111 ],[ -587, -111 ],[ 1166, 0 ]]]),
(-83, 83, [[[ -37//2, -3//2 ],[ 37//2, -3//2 ],[ 5//2, -5//2 ],[ 21, 1 ]],[[ -37//2, 1//2 ],[ -7//2, -3//2 ],[ -6, -2 ],[ 27//2, -1//2 ]],[[ 17//2, -1//2 ],[ 3, 1 ],[ 11//2, 1//2 ],[ -11//2, 1//2 ]],[[ 5, 0 ],[ -1//2, 1//2 ],[ 1//2, 1//2 ],[ -4, 0 ]],[[ -15, 0 ],[ 7//2, -3//2 ],[ -17//2, -3//2 ],[ 29//2, -1//2 ]],[[ 12, 1 ],[ -25//2, 3//2 ],[ -5//2, 3//2 ],[ -33//2, -1//2 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]]]),
(-340, 85, [[[ -67, 0 ],[ 16, -7 ],[ -16, -7 ],[ 66, 0 ]],[[ 9, 0 ],[ -2, 1 ],[ 2, 1 ],[ -10, 0 ]],[[ 145, -1 ],[ -19, 16 ],[ 36, 15 ],[ -146, 1 ]],[[ -10, 0 ],[ 4, -1 ],[ -4, -1 ],[ 10, 0 ]],[[ -16, -1 ],[ 16, -1 ],[ 0, -2 ],[ 16, 1 ]],[[ 19, 0 ],[ -1, 2 ],[ 1, 2 ],[ -18, 0 ]],[[ -526, 9 ],[ 40, -57 ],[ -204, -54 ],[ 526, -9 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -38, 0 ],[ 11, -4 ],[ -11, -4 ],[ 39, 0 ]],[[ -114, -4 ],[ 52, -11 ],[ 10, -13 ],[ 111, 3 ]],[[ -124, -15 ],[ 154, -11 ],[ 108, -17 ],[ 138, 14 ]],[[ -213, -15 ],[ 183, -20 ],[ 90, -25 ],[ 212, 15 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -92, -15 ],[ 158, -7 ],[ 137, -10 ],[ 65, 17 ]],[[ -4, -39 ],[ 352, 5 ],[ 360, -6 ],[ 4, 39 ]],[[ -46, 0 ],[ 6, -5 ],[ -6, -5 ],[ 47, 0 ]],[[ 2000, 115 ],[ -1454, 195 ],[ -644, 230 ],[ -2000, -115 ]],[[ -85, 8 ],[ -57, -11 ],[ -88, -8 ],[ 93, -8 ]],[[ 277, -27 ],[ 176, 37 ],[ 289, 24 ],[ -296, 25 ]],[[ -171, -12 ],[ 157, -14 ],[ 77, -20 ],[ 152, 14 ]],[[ -172, 17 ],[ -127, -22 ],[ -180, -15 ],[ 173, -17 ]],[[ 281, 231 ],[ -2191, -14 ],[ -1995, 72 ],[ -281, -231 ]],[[ -457, -13 ],[ 212, -47 ],[ 27, -50 ],[ 456, 13 ]]]),
(-344, 86, [[[ -91, -17 ],[ -91, 17 ],[ -166, -1 ],[ 75, 16 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -77, 22 ],[ 219, -7 ],[ 126, 20 ],[ 75, -24 ]],[[ 43, 3 ],[ 2, -6 ],[ 21, -2 ],[ -29, -1 ]],[[ -253, -15 ],[ 40, 33 ],[ -239, 13 ],[ 255, 14 ]],[[ 52, -9 ],[ -97, 0 ],[ -53, -10 ],[ -50, 10 ]],[[ -99, -19 ],[ -85, 19 ],[ -222, 0 ],[ 123, 19 ]],[[ 187, 16 ],[ 13, -25 ],[ 226, -6 ],[ -171, -16 ]],[[ -189, -32 ],[ -116, 36 ],[ -362, -2 ],[ 211, 32 ]],[[ -187, 5 ],[ 149, 16 ],[ -54, 22 ],[ 221, -7 ]],[[ -161, 1 ],[ 93, 14 ],[ -88, 15 ],[ 163, 0 ]],[[ 371, -16 ],[ -360, -26 ],[ 61, -39 ],[ -361, 18 ]],[[ 158, 7 ],[ -45, -19 ],[ 149, -9 ],[ -166, -8 ]],[[ 59, -20 ],[ -200, 7 ],[ -126, -16 ],[ -51, 22 ]],[[ -58, 30 ],[ 285, -13 ],[ 207, 21 ],[ 46, -33 ]],[[ -65, 17 ],[ 185, -3 ],[ 118, 16 ],[ 73, -21 ]],[[ 101, -1 ],[ -70, -8 ],[ 23, -8 ],[ -73, 3 ]],[[ -21, -5 ],[ -33, 5 ],[ -48, 0 ],[ 27, 5 ]],[[ 212, -12 ],[ -229, -13 ],[ 45, -25 ],[ -234, 11 ]],[[ -86, 12 ],[ 151, 1 ],[ 41, 15 ],[ 99, -13 ]],[[ 83, 36 ],[ 247, -30 ],[ 298, 12 ],[ -83, -36 ]],[[ -101, 12 ],[ 160, 2 ],[ 48, 12 ],[ 67, -12 ]],[[ -1, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 0, 47 ],[ 343, -29 ],[ 343, 29 ],[ 0, -47 ]]]),
(-87, 87, [[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ -1, 0 ],[ 2, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ -5//2, 1//2 ],[ -6, -1 ],[ -5, 0 ],[ 5, -1 ]],[[ -7//2, 1//2 ],[ -4, -1 ],[ -3, 0 ],[ 5//2, -1//2 ]],[[ 13//2, -1//2 ],[ 0, 1 ],[ 5, 0 ],[ -7//2, 1//2 ]],[[ -27//2, -5//2 ],[ 53//2, 1//2 ],[ 13//2, -3//2 ],[ 13//2, 3//2 ]],[[ 39//2, -3//2 ],[ -11//2, 5//2 ],[ 25//2, 1//2 ],[ -25//2, 1//2 ]],[[ -28, -3 ],[ 56, 0 ],[ 0, -3 ],[ 28, 3 ]],[[ 13//2, -5//2 ],[ 20, 3 ],[ 43//2, -1//2 ],[ -14, 3 ]],[[ -8, 1 ],[ -3, -2 ],[ -8, 0 ],[ 8, -1 ]]]),
(-356, 89, [[[ -90, -17 ],[ 184, 0 ],[ 66, -10 ],[ 50, 11 ]],[[ 440, 13 ],[ -255, 38 ],[ 108, 47 ],[ -437, -5 ]],[[ -1, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ 462, -8 ],[ -143, 46 ],[ 275, 36 ],[ -418, 9 ]],[[ 0, -1 ],[ 9, 0 ],[ 10, 0 ],[ 0, 1 ]],[[ 88, 9 ],[ -113, 4 ],[ -54, 9 ],[ -49, -9 ]],[[ 7, 0 ],[ -6, 1 ],[ -1, 1 ],[ -12, -1 ]],[[ -33, 0 ],[ 16, -3 ],[ -16, -3 ],[ 32, 0 ]],[[ 107, -24 ],[ 144, 24 ],[ 232, -3 ],[ -107, 24 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ -215, 16 ],[ -64, -27 ],[ -239, -13 ],[ 200, -19 ]],[[ 62, -231 ],[ 1856, 110 ],[ 1928, -107 ],[ 2, 225 ]],[[ 7, 2 ],[ -28, 0 ],[ -11, 1 ],[ -7, -2 ]],[[ 785, 48 ],[ -768, 51 ],[ -32, 96 ],[ -785, -48 ]],[[ 211, -8 ],[ -30, 23 ],[ 156, 14 ],[ -185, 8 ]],[[ 4, 3 ],[ -27, -1 ],[ -27, 1 ],[ 4, -3 ]],[[ -19, 0 ],[ 9, -3 ],[ -6, -2 ],[ 31, 0 ]],[[ -43, -4 ],[ 54, -2 ],[ 11, -2 ],[ 11, 2 ]],[[ 180, -13 ],[ 19, 23 ],[ 161, 10 ],[ -161, 10 ]],[[ -12, -29 ],[ 242, 12 ],[ 139, -5 ],[ -22, 15 ]],[[ 205, -80 ],[ 574, 65 ],[ 761, -20 ],[ -257, 85 ]],[[ -79, 6 ],[ -7, -11 ],[ -59, -3 ],[ 59, -4 ]],[[ -74, 34 ],[ -272, -25 ],[ -326, 9 ],[ 85, -38 ]],[[ 543, 130 ],[ -1306, -9 ],[ -856, 114 ],[ -543, -130 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -15, 0 ],[ 6, -2 ],[ -3, -1 ],[ 13, 0 ]]]),
(-91, 91, [[[ 9//2, 1//2 ],[ 2, -1 ],[ 4, 0 ],[ -7//2, -1//2 ]],[[ 11//2, -1//2 ],[ -19//2, -1//2 ],[ -3//2, -1//2 ],[ -11//2, 1//2 ]],[[ 123//2, -3//2 ],[ -125//2, -15//2 ],[ 17//2, -9//2 ],[ -63, 2 ]],[[ 23//2, 1//2 ],[ -3, -2 ],[ 13//2, -1//2 ],[ -23//2, -1//2 ]],[[ -677//2, 57//2 ],[ 606, 27 ],[ 105//2, 59//2 ],[ 677//2, -57//2 ]],[[ -19//2, -7//2 ],[ -85//2, 9//2 ],[ -45//2, -1//2 ],[ 12, 4 ]],[[ 1, 0 ],[ 0, 0 ],[ 1, 0 ],[ -1, 0 ]],[[ 6, 0 ],[ -5//2, -1//2 ],[ 5//2, -1//2 ],[ -5, 0 ]]]),
(-372, 93, [[[ 27, 3 ],[ 11, -4 ],[ 28, -1 ],[ -23, -2 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ 19, 0 ],[ -14, -2 ],[ 7, -1 ],[ -15, 0 ]],[[ -68, -7 ],[ 1, 14 ],[ -68, 0 ],[ 68, 7 ]],[[ 22, -3 ],[ -44, 0 ],[ -8, -3 ],[ -22, 3 ]],[[ 1520, 91 ],[ 0, -182 ],[ 1520, -91 ],[ -1520, -91 ]],[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ -6, -1 ],[ -7, 1 ],[ -10, 0 ],[ 4, 1 ]],[[ 80, -9 ],[ -156, -2 ],[ -18, -9 ],[ -80, 9 ]],[[ 200, 3 ],[ -122, -22 ],[ 111, -16 ],[ -230, 1 ]],[[ 6, -4 ],[ -47, 2 ],[ -23, -2 ],[ -6, 4 ]],[[ -87, 6 ],[ 95, 5 ],[ -11, 8 ],[ 74, -3 ]],[[ -47, 6 ],[ 94, 0 ],[ 24, 7 ],[ 47, -8 ]],[[ 4354, -222 ],[ -4677, -280 ],[ 579, -448 ],[ -4354, 222 ]],[[ -80, 28 ],[ 278, -7 ],[ 190, 19 ],[ 64, -27 ]],[[ -1, 0 ],[ 2, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ -267, -20 ],[ -14, 38 ],[ -298, 14 ],[ 309, 20 ]],[[ 106, -5 ],[ -120, -7 ],[ 20, -10 ],[ -106, 5 ]],[[ -150, -27 ],[ -152, 27 ],[ -299, 0 ],[ 149, 27 ]],[[ 281, -3 ],[ -252, -27 ],[ 123, -20 ],[ -292, 3 ]],[[ -39, -4 ],[ -7, 7 ],[ -40, 0 ],[ 31, 4 ]],[[ -91, 30 ],[ 350, -14 ],[ 130, 22 ],[ 105, -30 ]],[[ 17, -2 ],[ -30, 0 ],[ -5, -2 ],[ -13, 2 ]],[[ 172, -5 ],[ -128, -13 ],[ 50, -13 ],[ -134, 2 ]],[[ 43, 8 ],[ 51, -8 ],[ 83, 0 ],[ -40, -8 ]]]),
(-376, 94, [[[ 33, -4 ],[ 17, 3 ],[ 50, 1 ],[ -17, 3 ]],[[ 97, 29 ],[ -194, 0 ],[ -359, 29 ],[ -97, -29 ]],[[ -12, 5 ],[ -49, -1 ],[ -49, 1 ],[ -12, -5 ]],[[ 7, 1 ],[ -12, 1 ],[ -8, 1 ],[ -13, -1 ]],[[ 37, 0 ],[ -6, 2 ],[ 21, 7 ],[ -39, 0 ]],[[ 9, 0 ],[ -2, 1 ],[ 2, 1 ],[ -11, 0 ]],[[ 313, 16 ],[ -230, 19 ],[ -108, 66 ],[ -451, -32 ]],[[ -2351, 140 ],[ 0, -280 ],[ -2351, -140 ],[ 2351, -140 ]],[[ 0, 0 ],[ -1, 0 ],[ -1, 0 ],[ 0, 0 ]],[[ -5, -1 ],[ 10, 0 ],[ 7, -1 ],[ 5, 1 ]],[[ 7, 3 ],[ -14, 0 ],[ -57, 3 ],[ -7, -3 ]],[[ 0, 0 ],[ -1, 0 ],[ 1, 0 ],[ 0, 0 ]],[[ -72, -5 ],[ 35, -3 ],[ 13, -9 ],[ 42, 2 ]],[[ -153, 25 ],[ -131, -21 ],[ -284, -4 ],[ 131, -21 ]],[[ -9, 0 ],[ -1, -1 ],[ -7, -2 ],[ 20, -1 ]],[[ 87, -16 ],[ 68, 7 ],[ 320, 8 ],[ -87, 16 ]],[[ 26, -11 ],[ 51, 3 ],[ 183, -2 ],[ -17, 10 ]],[[ 1, 1 ],[ -4, 0 ],[ -23, 1 ],[ -3, -1 ]],[[ 5, 1 ],[ -6, 0 ],[ -15, 1 ],[ -1, -1 ]],[[ -68, -4 ],[ 29, -3 ],[ 39, -7 ],[ 29, 3 ]],[[ 9, 0 ],[ 0, 1 ],[ 9, 2 ],[ -21, 1 ]],[[ 149, 11 ],[ -149, 11 ],[ -40, 22 ],[ -189, -11 ]],[[ 95, 3 ],[ -27, 4 ],[ 16, 12 ],[ -55, -1 ]],[[ -84, -3 ],[ 67, -6 ],[ 17, -9 ],[ 67, 6 ]]]),
(-95, 95, [[[ 0, 0 ],[ 1, 0 ],[ -1, 0 ],[ 1, 0 ]],[[ 7//2, -1//2 ],[ -7, 0 ],[ -3//2, -1//2 ],[ -7//2, 1//2 ]],[[ 15, 0 ],[ -12, -1 ],[ 12, -1 ],[ -16, 0 ]],[[ 3//2, 1//2 ],[ -1//2, -1//2 ],[ 1, 0 ],[ -1, 0 ]],[[ 15//2, 3//2 ],[ 11//2, -3//2 ],[ 7, 0 ],[ -9//2, -1//2 ]],[[ 17//2, 1//2 ],[ 0, -1 ],[ 17//2, -1//2 ],[ -17//2, -1//2 ]],[[ -1, 0 ],[ 1, 0 ],[ 0, 0 ],[ 1, 0 ]],[[ 25//2, 1//2 ],[ -6, -1 ],[ 13//2, -1//2 ],[ -7, 0 ]],[[ 9//2, -1//2 ],[ -9, 0 ],[ -1//2, -1//2 ],[ -9//2, 1//2 ]],[[ -14, 0 ],[ 10, 1 ],[ -10, 1 ],[ 14, 0 ]],[[ 21//2, 1//2 ],[ -3//2, -5//2 ],[ 21//2, -1//2 ],[ -39//2, -3//2 ]],[[ -27//2, 1//2 ],[ 23//2, 1//2 ],[ -15//2, 3//2 ],[ 27//2, -1//2 ]]]),
(-388, 97, [[[ -4, 1 ],[ -10, -1 ],[ -8, 0 ],[ 4, -1 ]],[[ 42, -1 ],[ -10, 4 ],[ 22, 2 ],[ -26, 1 ]],[[ 11, 0 ],[ -5, 1 ],[ 6, -1 ],[ 6, 1 ]],[[ 0, -3 ],[ 34, 2 ],[ 20, -1 ],[ -2, 3 ]],[[ -165, -2 ],[ 89, -13 ],[ -60, -16 ],[ 157, 2 ]],[[ 20, 10 ],[ -126, -5 ],[ -64, 5 ],[ -9, -11 ]],[[ -6, -37 ],[ 399, 17 ],[ 279, -13 ],[ 6, 37 ]],[[ 483, 88 ],[ -984, 13 ],[ -511, 67 ],[ -424, -73 ]],[[ 108, -5 ],[ -5, 11 ],[ 103, 6 ],[ -91, 6 ]],[[ 6, 1 ],[ -11, 0 ],[ -6, 1 ],[ -5, -1 ]],[[ 1166, -18 ],[ -371, 106 ],[ 753, 102 ],[ -1166, 18 ]],[[ -90, 17 ],[ -129, -18 ],[ -199, -2 ],[ 120, -20 ]],[[ -732, 99 ],[ -476, -106 ],[ -1282, -19 ],[ 732, -99 ]],[[ 26, -1 ],[ -6, 4 ],[ 19, 2 ],[ -38, 1 ]],[[ 53, 8 ],[ -129, 3 ],[ -53, 8 ],[ -76, -11 ]],[[ 40, -3 ],[ 20, 7 ],[ 40, 3 ],[ -60, 4 ]],[[ 24, -17 ],[ 160, 11 ],[ 163, -4 ],[ -41, 19 ]],[[ 26, -7 ],[ 57, 8 ],[ 74, -1 ],[ -43, 9 ]],[[ 188, -49 ],[ 399, 44 ],[ 376, 0 ],[ -188, 39 ]],[[ -255, -18 ],[ 313, -20 ],[ 53, -26 ],[ 256, 18 ]],[[ -230, -52 ],[ 567, -2 ],[ 365, -48 ],[ 303, 53 ]],[[ -9, 2 ],[ -17, -2 ],[ -18, 0 ],[ 9, -2 ]],[[ 3, -1 ],[ 12, 1 ],[ 7, 0 ],[ -4, 1 ]]])
]
################################################################################
#
#  Orbit stabilizer
#
################################################################################

function _orbit_stabilizer(G, idity, a)
  OT = Tuple{typeof(idity), FakeFmpqMat}[(idity, hnf(basis_matrix(FakeFmpqMat, a)))]
  Y = typeof(idity)[]
  m = 1
  while m <= length(OT)
    b = OT[m][2]
    for g in G
      c = _operate(g, b)
      i = findfirst(x -> x[2] == c, OT)
      if i isa Nothing
        push!(OT, (OT[m][1] * g, c))
      elseif i isa Int
        w = OT[m][1] * g * inv(OT[i][1])
        if !(w in Y || inv(w) in Y)
          push!(Y, w)
        end
      end
    end
    m += 1
  end

  return OT, Y
end

function _operate(g::AbstractAssociativeAlgebraElem, b)
  M = representation_matrix(g, :right)
  c = hnf(b * FakeFmpqMat(M))
  return c
end
